---------------------------------------------------------------------------
-- This VHDL file was developed by Daniel Llamocca (2018).  It may be
-- freely copied and/or distributed at no cost.  Any persons using this
-- file for any purpose do so at their own risk, and are responsible for
-- the results of such use.  Daniel Llamocca does not guarantee that
-- this file is complete, correct, or fit for any particular purpose.
-- NO WARRANTY OF ANY KIND IS EXPRESSED OR IMPLIED.  This notice must
-- accompany any copy of this file.
--------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.std_logic_arith.all;

-- This ALU treats data as signed numbers. This is very important, especially for subtraction
--    the result is a signed number.
-- We include operations that have carry in or borrow for summation and subtraction
-- Instead of including the cin or bin input, we instead encode it in 'sel'.
-- When sel = 011 -> A+B+cin (where cin=1) , sel = 100 -> A-B-bin (where bin=1)
entity alu_arith is
	generic (N: INTEGER:= 8);
	port (A, B: in std_logic_vector (N-1 downto 0);
	      sel: in std_logic_vector (2 downto 0);
	      y: out std_logic_vector (N-1 downto 0);
	      V,C: out std_logic);
end alu_arith;

architecture structure of alu_arith is

    component super_addsub -- IMPORTANT: Data is treated as signed (2C)
        generic (N: INTEGER:= 4;
                 DIRECTION: STRING:= "SUB"; -- "UNUSED", "ADD", "SUB"
                 cin_used: STRING:= "NO"); -- "YES", "NO"
        port( addsub   : in std_logic; -- add_sub = 1 -> SUBTRACTION, add_sub = 0 -> ADDITION. This only applies if DIRECTION = "UNUSED"
              cin      : in std_logic; -- ADDITION: active-high carry in
                                         -- SUBTRACTION: active-low borrow in: if cin = 1 -> X-Y, if cin = 0 -> X-Y-1
              x, y     : in std_logic_vector (N-1 downto 0);
              s        : out std_logic_vector (N-1 downto 0);
              overflow : out std_logic; -- This is always cn XOR cn-1, i.e., the overflow when numbers are treated as 2C
              cout     : out std_logic);
    end component;

   signal Ap1, Am1, Bm1, ApB, AmB, ApBc, AmBc: std_logic_vector (N-1 downto 0);
   signal my_one: std_logic_vector (N-1 downto 0);
   
   signal VB, CB: std_logic_vector (7 downto 0); -- 8 operations
	
begin

my_one(N-1 downto 1) <= (others => '0'); my_one(0) <= '1'; -- my_one = 1

-- A + 1
f1: super_addsub generic map (N => N, DIRECTION => "ADD", cin_used => "NO")
    port map (addsub => '0', cin => '0', x => A, y => my_one , s => Ap1, overflow => VB(1), cout => CB(1));

-- A + B
f2: super_addsub generic map (N => N, DIRECTION => "ADD", cin_used => "NO")
    port map (addsub => '0', cin => '0', x => A, y => B , s => ApB,overflow => VB(2), cout => CB(2));

-- A + B + 1. This is when there is a carry in.
f3: super_addsub generic map (N => N, DIRECTION => "ADD", cin_used => "YES")
    port map (addsub => '0', cin => '1', x => A, y => my_one , s => ApBc,overflow => VB(3), cout => CB(3));

-- A - B -1 = A + not(B). A and B are treated as signed numbers
f4: super_addsub generic map (N => N, DIRECTION => "SUB", cin_used => "YES")
    port map (addsub => '0', cin => '0', x => A, y => B , s => AmBc, overflow => VB(4), cout => CB(4)); -- cin =0 -> bin=1
    
-- A - B = A + not(B) + 1. A and B are treated as signed numbers
f5: super_addsub generic map (N => N, DIRECTION => "SUB", cin_used => "NO")
    port map (addsub => '0', cin => '0', x => A, y => B , s => AmB, overflow => VB(5), cout => CB(5));

-- A - 1
f6: super_addsub generic map (N => N, DIRECTION => "SUB", cin_used => "NO")
    port map (addsub => '0', cin => '0', x => A, y => my_one , s => Am1, overflow => VB(6), cout => CB(6));
    
V <= VB (conv_integer(unsigned(sel)));
C <= CB (conv_integer(unsigned(sel)));

VB(0) <= '0'; CB(0) <= '0';
--VB(1) <= '0'; CB(0) <= '0'; 
--VB(2) <= '0'; CB(0) <= '0';
--VB(3) <= '0'; CB(0) <= '0';
--VB(4) <= '0'; CB(0) <= '0';
--VB(5) <= '0'; CB(0) <= '0';
--VB(6) <= '0'; CB(0) <= '0';
VB(7) <= '0'; CB(7) <= '0';
	 
-- Multiplexor	 
	 with sel select
			y <= A   when "000",
			     Ap1 when "001",
				 ApB when "010",
				 ApBc when "011",
				 AmBc when "100", -- A + not(B) = A - B - 1
				 AmB when "101", -- A + not(B) + 1 = A-B
				 Am1 when "110",
				 B when "111",
				 (others => '0') when others;

end structure;

