---------------------------------------------------------------------------
-- This VHDL file was developed by Daniel Llamocca (2017).  It may be
-- freely copied and/or distributed at no cost.  Any persons using this
-- file for any purpose do so at their own risk, and are responsible for
-- the results of such use.  Daniel Llamocca does not guarantee that
-- this file is complete, correct, or fit for any particular purpose.
-- NO WARRANTY OF ANY KIND IS EXPRESSED OR IMPLIED.  This notice must
-- accompany any copy of this file.
--------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all; -- treats all operations here as unsigned.

-- The condition on branching instructions: we should do it with one signal coming out of the Instruction Decoder
-- For now, we encode it here.
entity instload_ctrl is
    generic (H: integer:= 8; -- Address word size
             N: integer:= 16); -- Instruction word size. H <= N
	port ( resetn, clock: in std_logic;
		   step: in std_logic; -- only when it is one this works
		   start: in std_logic; -- a pulse is detected
		   L_in, L_ex: std_logic; -- a pulse is detected
		   we_ex: in std_logic; -- can be detected continuously. If connected to a switch, we need a pulse generator of 1 clock cycle.
		   D_ex : in std_logic_vector (N-1 downto 0);
		   
		   isbranch: in std_logic;		   
		   PC: in std_logic_vector (H-1 downto 0);
		   
		   -- Outputs:
		   E_PC, sclr_PC: out std_logic;
		   stop: out std_logic; -- This is for the Instruction Decoder so that all its outputs are 0.
		   IM_we: out std_logic;
		   IM_D: out std_logic_vector (N-1 downto 0));
end instload_ctrl;

architecture structure of instload_ctrl is
	
	type state is (S1, S2, S3, S4, S2x, S3x, S4x);
	signal y: state;
	
	type chunk is array (2**H -1 downto 0) of std_logic_vector (N-1 downto 0);
	
	signal CD: chunk;
	signal i, T: integer range 0 to 2**H - 1; -- H should not be more than 32 bits.
		
begin

a0: assert (H <= N)
    report "H cannot be larger than N!!!"
	severity error;

ga: if (H /= 6 or N/=16) and (H/=4 or N/=8) generate
        gi: for i in 0 to 2**H - 1 generate
                CD(i) <= (others => '0');
            end generate;
    end generate;
    
-- Filling up a initial program: We can also do this by reading a text file    
g_Morris: if H = 6 and N = 16 generate
                CD (0)  <= "1001100011---100"; -- ldi r3, 4
                CD (1)  <= "1001100000---010"; -- ldi r0, 2
                CD (2)  <= "1001100001---111"; -- ldi r1,7
                CD (3)  <= "1000010001001100"; -- adi R1, R1, 4
                
                CD (4)  <= "1000010000000001"; -- adi r0, r0, 1
                CD (5)  <= "0000110001001---"; -- dec r1, r1
                CD (6)  <= "1100000111001011"; -- brz r1, start
                CD (7)  <= "1110000---011---"; -- jpm r3, loop
        
            np: for i in 8 to 2**H - 1 generate
                    CD(i) <= (others => '0');
                end generate;                
          
    end generate;

g_vbc: if H = 4 and N = 8 generate
            -- 2 to 13
--            CD (0)  <= "00100010"; -- loadi R0,2
--            CD (1)  <= "00110100"; -- loadi R1,4
--            CD (2)  <= "11000000"; -- out R0
--            CD (3)  <= "01100001"; -- addi R0,1
--            CD (4)  <= "01110001"; -- addi R1,1
--            CD (5)  <= "11110010"; -- jnz R1, loop
--            CD (6)  <= "00100001"; -- load R0,1
--            CD (7)  <= "11100000"; -- jnz R0, start

            -- 4 to 11
            CD (0)  <= "00100100"; -- loadi R0,4
            CD (1)  <= "00111000"; -- loadi R1,8
            CD (2)  <= "11000000"; -- out R0
            CD (3)  <= "01100001"; -- addi R0,1
            CD (4)  <= "01110001"; -- addi R1,1
            CD (5)  <= "11110010"; -- jnz R1, loop
            CD (6)  <= "00110001"; -- load R1,1
            CD (7)  <= "11110000"; -- jnz R1, start
            
            np: for i in 8 to 2**H - 1 generate
                    CD(i) <= (others => '0');
                end generate; 
       end generate;
                
-- Main FSM:
	Transitions: process (resetn, clock, start, i, L_in, L_ex, we_ex, isbranch, PC, T, step)
	begin
		if resetn = '0' then -- asynchronous signal
			y <= S1; i <= 0; T <= 0; -- if resetn asserted, go to initial state: S1			
		elsif (clock'event and clock = '1') then
			case y is
				when S1 =>
				    T <= (2**H - 1);
					if start = '1' then
					   y <= S2;
					else
					   if L_in = '1' then
					       y <= S3;
					   elsif L_ex  = '1' then
					       y <= S4;
					   else
					       y <= S1;
					   end if;
					end if;
				
				when S2 =>
				    if step = '1' then
                        if PC = T then
                            --if IR(N-1 downto N-7) = "1100000" or  IR(N-1 downto N-7) = "1100001" or IR(N-1 downto N-7) = "1110000" then
                            if isbranch = '1' then
                                y <= S2;
                            else
                                y <= S2x;
                            end if;
                        else
                            y <= S2;
                        end if;
                    else
                        y <= S2;
                    end if;
                    
                when S2x =>
                    if start = '1' then y <= S2x; else y <= S1; end if;
				    
				when S3 =>
				    if i = (2**H - 1) then
				        y <= S3x; i <= 0;
				    else
				        y <= S3; i <= i + 1;
				    end if;
				    
                when S3x =>
                    if start = '1' then y <= S3x; else y <= S1; end if;
                    				    			
				when S4 =>
				    if we_ex = '1' then
                        if i = (2**H - 1) then
                            y <= S4x; i <= 0;
                        else
                            if start = '1' then y <= S2; i <= 0; T <= i; else i <= i + 1; y <= S4; end if;
                        end if;
                    else
                        y <= S4;
                    end if;
                    
                when S4x =>
                    if start = '1' then y <= S4x; else y <= S1; end if;                    
					
			end case;			
		end if;		
	end process;
	
	Outputs: process (y, start, i, CD, we_ex, D_ex, step)
	begin
		-- Initialization of FSM outputs:
		IM_we <= '0'; IM_D <= (others => '0'); E_PC <= '0'; sclr_PC <= '0';
		stop <= '0';
		
		case y is
			when S1 =>
			    E_PC <= '1'; sclr_PC <= '1'; stop <= '1';
			    
			when S2 =>
			    if step = '1' then
			       E_PC <= '1';
			    else
			       stop <= '1'; -- if step = '0', everything freezes, the PC also freezes.
			    end if;
			
			when S2x =>
			    stop <= '1';
			    
			when S3 =>
			    E_PC <= '1';
			    IM_we <= '1';
			    IM_D <= CD(i);
			    stop <= '1';
			
			when S3x =>
			    stop <= '1';
			    
			when S4 =>
                stop <= '1';
                if we_ex = '1' then
                    IM_D <= D_ex; IM_we <= '1'; E_PC <= '1';
                    if i /= 2**H -1 then
                        if start = '1' then E_PC <= '1'; sclr_PC <= '1'; end if;
                    end if;
                end if;
            
            when S4x =>
                stop <= '1';
			
		end case;
	end process;
 
end structure;
